/*******************************************************************************
  This file is part of the kde-game 'kjumpingCube'

  Copyright (C) 1998 by Matthias Kiefer <Matthias.Kiefer@stud.uni-karlsruhe.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  *******************************************************************************/  

#include <kapp.h>
#include <kmsgbox.h>
#include <qpalette.h>
#include <klocale.h>
#include <kconfig.h>
#include <kconfigbase.h>
#include <qtimer.h>

#include "cubebox.h"

#define MAX_CUBESIZE 96
#define MIN_CUBESIZE 36
#define DEFAULT_CUBESIZE 60

CubeBox::CubeBox(int dim,int zoom, QPalette *color1, QPalette *color2,
		 QWidget *parent, const char *name) : QWidget(parent,name)
{
  int i,j;

  actualPlayer=1;
  doingMove=FALSE;
  stopped=FALSE;
  computerPlayer[0]=FALSE;
  computerPlayer[1]=FALSE;
      
  // size of playfield
  cubeDim=dim;
  // size of cube
  cubeSize=DEFAULT_CUBESIZE/5*zoom;
    
  // create cubes
  cubes=new Cube**[cubeDim];
  for(i=0;i<cubeDim;i++)
    cubes[i]=new Cube*[cubeDim];
  for(i=0;i<cubeDim;i++)
    for(j=0;j<cubeDim;j++)
      {
	cubes[i][j]=new Cube(this,i,j);
      }


  // set maxValues 

   for(i=0;i<cubeDim;i++)
    for(j=0;j<cubeDim;j++)
      {
	if(i==0 || i==(cubeDim-1))
	  if(j==0 || j==(cubeDim-1))
	      cubes[i][j]->setMaxValue(2); // maxValue of the cornercubes
	  else cubes[i][j]->setMaxValue(3);  // left and right egde
	else if(j==0 || j==(cubeDim-1))
	  cubes[i][j]->setMaxValue(3);  // upper and lower edge;
	else cubes[i][j]->setMaxValue(4); // remaining
      }

   for(i=0;i<cubeDim;i++)
     for(j=0;j<cubeDim;j++)
     {  
       // when colors changed, show new colors
       connect(this,SIGNAL(colorChanged()),cubes[i][j],SLOT(updateColor()));
       
       // get clicks
       connect(cubes[i][j],SIGNAL(cubeClicked(int,int)),SLOT(checkClick(int,int)));
     }

    // dont' connect here, because then checkComputerplayer(int) is called first
    // and not TopWindow::changePlayer 
   //connect(this,SIGNAL(playerChanged(int)),SLOT(checkComputerplayer(int)));
     
   // set colors and sizes

   // set new Colors
   Cube::setColors(color1,color2); // colors are static in Cubes
   
   // set new size
   setFixedSize(cubeDim*cubeSize,cubeDim*cubeSize);
   
   
   previousField.copy(this);
//   think.init(4);
}


CubeBox::~CubeBox()
{
  if(isActive())
     stopActivities();
  
  
  int i,j;
  for(i=0;i<cubeDim;i++)
    for(j=0;j<cubeDim;j++)
      {
	delete cubes[i][j];
      }     
  
  for(i=0;i<cubeDim;i++)
    delete [] cubes[i];
  
  delete [] cubes;
   
}

// Slots

void CubeBox::showWinner(int winner)
{
  QString s;
  s.sprintf(klocale->translate("Winner is Player %d"),winner);
  KMsgBox::message(this,klocale->translate("Winner"),s);
}


void CubeBox::setZoom(int zoom)
{
  // set new cubesize
  cubeSize=DEFAULT_CUBESIZE/5*zoom;
  setFixedSize(cubeDim*cubeSize,cubeDim*cubeSize);
}


int CubeBox::zoomIn()
{
  if(cubeSize < MAX_CUBESIZE)
    {
      cubeSize+=DEFAULT_CUBESIZE/10;
      setFixedSize(cubeDim*cubeSize,cubeDim*cubeSize);
    }

  return cubeSize*5/DEFAULT_CUBESIZE;
}

int CubeBox::zoomOut()
{
  if(cubeSize > MIN_CUBESIZE)
    {
      cubeSize-=DEFAULT_CUBESIZE/10;
      setFixedSize(cubeDim*cubeSize,cubeDim*cubeSize);
    }
  
  return cubeSize*5/DEFAULT_CUBESIZE;


}


bool CubeBox::maxZoom()
{
  if(cubeSize>=MAX_CUBESIZE)
    return TRUE;
  else return FALSE;
}

bool CubeBox::minZoom()
{
  if(cubeSize<=MIN_CUBESIZE)
    return TRUE;
  else return FALSE;
}

int CubeBox::zoom()
{
  return cubeSize*5/DEFAULT_CUBESIZE;
}

int CubeBox::dim()
{
  return cubeDim;
}

void CubeBox::setColors(QPalette *color1,QPalette *color2)
{
  Cube::setColors(color1,color2);
  emit colorChanged();
}

int CubeBox::setSkill(int skill)
{
   return think.setSkill(skill);
}

int CubeBox::skill()
{
   return think.getSkill();
}


void CubeBox::setComputerplayer(int player,bool flag)
{
   if(player < 1 || player > 2)
      return;
   
   computerPlayer[player-1]=flag;

}

bool CubeBox::isComputer(int player)
{
   return computerPlayer[player-1];
}

void CubeBox::checkComputerplayer(int player)
{
   // checking if a process is running or the Widget isn't shown yet
   if(isActive() || !isVisibleToTLW())  
      return;
   
   stopped=FALSE;   
   if(computerPlayer[player-1] && player==actualPlayer)
   {  
      Cube::allowClicks(FALSE);
      int row=0,column=0;
      CubeSimBox field(this);
      
      emit startedThinking();
      think.getHint(row,column,player,field);
      emit stoppedThinking();
      
      if(stopped)
      {
	 return;
      }
      
      cubes[row][column]->showTip();

      QTimer::singleShot(1000,cubes[row][column],SLOT(hideTip()));
      QString s;
      s.sprintf("row: %d , column: %d",row+1,column+1);
      debug(s);  
      doMove(row,column);
      
      Cube::allowClicks(TRUE);
   }
      
}

bool CubeBox::isActive()
{
   if(think.isActive() || doingMove)
      return TRUE;
   
   return FALSE;
}

void CubeBox::stopActivities()
{
   stopped=TRUE;
   stopThinking();
   stopLoop();
}


void CubeBox::checkClick(int row,int column)
{
   // if a move hasn't finished yet don't do another move
   if(doingMove)
      return;
   
   if(actualPlayer==cubes[row][column]->getOwner() ||
		   cubes[row][column]->getOwner()==0)
      doMove(row,column);
   
}		

void CubeBox::reset()
{
   int i,j;

   stopLoop();
   doingMove=FALSE;
   if(isActive())
      stopActivities();
   
   Cube::allowClicks(TRUE);
   Cube::playerPoints[0]=0;
   Cube::playerPoints[1]=0;
   actualPlayer=1;
   
   for(i=0;i<cubeDim;i++)
      for(j=0;j<cubeDim;j++)
	 cubes[i][j]->reset();
   
   emit playerChanged(1);

}
		


void CubeBox::doMove(int row,int column)
{
   // if a move hasn't finished yet don't do another move
   if(doingMove)
      return;
 
   stopped=FALSE;  
   doingMove=TRUE;
   
   // fuer undo-funktion Feld kopieren
   previousField.copy(this);
    
   Cube::allowClicks(FALSE);
   
   cubes[row][column]->increase(actualPlayer);

   startLoop();   
   
}

void CubeBox::startLoop()
{
   loop.row=0;
   loop.column=0;
   loop.finished=TRUE;
   
   startTimer(500);
   nextLoopstep();
}


void CubeBox::stopLoop()
{
   killTimers();
   doingMove=FALSE;
   Cube::allowClicks(TRUE);
}


void CubeBox::nextLoopstep()
{
   // search cube with to many points
   while(!cubes[loop.row][loop.column]->maxReached())
   {
      loop.column++;
      if(loop.column==cubeDim)
        {
          if(loop.row==cubeDim-1)
	  {
	     if(!loop.finished)
	     {
		loop.row=0;
		loop.column=0;
		loop.finished=TRUE;
		return;
	     }
	     else   // Schleife beendet
	     {
		stopLoop();
	   
		actualPlayer=(actualPlayer==1) ? 2 : 1;  // Spieler ndern
		emit playerChanged(actualPlayer);
		return;
	     } 
	  }
	  else
	  {
	     loop.row++;
	     loop.column=0;
	  }
       }
    }
    
 
   increaseNeighbours(loop.row,loop.column);
   cubes[loop.row][loop.column]->decrease();
   loop.finished=FALSE;
	       
    // check if someone has won
   if(Cube::playerPoints[actualPlayer-1]==cubeDim*cubeDim)
   {
      stopLoop();
      showWinner(actualPlayer);
      reset();
      return;
   }	
   
}

// this function does the next step of the loop for checking the playingfield
void CubeBox::timerEvent(QTimerEvent *)
{
   nextLoopstep();
}



void CubeBox::increaseNeighbours(int row, int column)
{
   if(row==0)
   {
    	if(column==0)  // linke obere Ecke
	{
	   cubes[0][1]->increase(actualPlayer);
	   cubes[1][0]->increase(actualPlayer);
	   return;
	}  
	else if(column==cubeDim-1)  // rechte obere Ecke
	{
	   cubes[0][cubeDim-2]->increase(actualPlayer);
	   cubes[1][cubeDim-1]->increase(actualPlayer);
	   return;
	}
	else  // oberer Rand
	{
	   cubes[0][column-1]->increase(actualPlayer);
	   cubes[0][column+1]->increase(actualPlayer);
	   cubes[1][column]->increase(actualPlayer);
	   return;
	}
   }
   else if(row==cubeDim-1) 
   {  
      if(column==0)  // linke untere Ecke
      {
         cubes[cubeDim-2][0]->increase(actualPlayer);
         cubes[cubeDim-1][1]->increase(actualPlayer);
         return;
      }
   
      else if(column==cubeDim-1) // rechte untere Ecke
      {
         cubes[cubeDim-2][cubeDim-1]->increase(actualPlayer);
         cubes[cubeDim-1][cubeDim-2]->increase(actualPlayer);
	 return;
      }
      else  // unterer Rand
      {
	cubes[cubeDim-1][column-1]->increase(actualPlayer);
	cubes[cubeDim-1][column+1]->increase(actualPlayer);
	cubes[cubeDim-2][column]->increase(actualPlayer);
	return; 
      }
   }
   else if(column==0) // linker Rand
   {
      cubes[row-1][0]->increase(actualPlayer);
      cubes[row+1][0]->increase(actualPlayer);
      cubes[row][1]->increase(actualPlayer);
      return;
   }
   else if(column==cubeDim-1)  // rechter Rand
   {
      cubes[row-1][cubeDim-1]->increase(actualPlayer);
      cubes[row+1][cubeDim-1]->increase(actualPlayer);
      cubes[row][cubeDim-2]->increase(actualPlayer);
      return;
   }
   else
   {
      cubes[row][column-1]->increase(actualPlayer);
      cubes[row][column+1]->increase(actualPlayer);
      cubes[row-1][column]->increase(actualPlayer);
      cubes[row+1][column]->increase(actualPlayer);
      return;
   }
}


bool CubeBox::undo(bool fromSessionManagment)
{
   // if a move hasn't finished yet don't do this
   if(isActive())
      return FALSE;
   
   int i,j;
   for(i=0;i<cubeDim;i++)
      for(j=0;j<cubeDim;j++)
      {
	 *cubes[i][j]=previousField.cube(i,j);
      }
      
      if(!fromSessionManagment)
      {
	 actualPlayer = (actualPlayer==1) ? 2 : 1;
         emit playerChanged(actualPlayer);
      }
      
      return TRUE;
}


void CubeBox::getHint()
{
   if(think.isActive() || doingMove)
      return;

   emit startedThinking();
      
   int row=0,column=0;
   CubeSimBox field(this);
   if(!think.getHint(row,column,actualPlayer,field))
      return;  // return if thinking was stopped

   cubes[row][column]->showTip();

   QTimer::singleShot(2000,cubes[row][column],SLOT(hideTip()));
   QString s;
   s.sprintf("row: %d , column: %d",row+1,column+1);
   debug(s);
   
   emit stoppedThinking();
  
}


void CubeBox::stopThinking()
{
   think.stop();
}
   
// Events


void CubeBox::resizeEvent(QResizeEvent *)
{
  // rearrange Cubes
  
  for(int i=0;i< cubeDim;i++)
    for(int j=0;j<cubeDim;j++)
      cubes[i][j]->setGeometry(j*cubeSize,i*cubeSize,cubeSize,cubeSize);
 
}


void CubeBox::saveGame(KConfig *config)
{
   stopped=TRUE;
   
   if(doingMove)
   {
      stopLoop();
      undo(TRUE);
   }
      
   if(think.isActive())
      stopThinking();
		      
  // save aktuell player
  config->writeEntry("onTurn",actualPlayer);
  
  QStrList list;
  list.setAutoDelete(TRUE);
  QString owner, value, key;
  
  for(int row=0; row < cubeDim ; row++)
    for(int column=0; column < cubeDim ; column++)
      {
	key.sprintf("%u,%u",row,column);
	owner.sprintf("%u",cubes[row][column]->getOwner());
	value.sprintf("%u",cubes[row][column]->value());
	list.append((const char *) owner);
	list.append((const char *) value);
	config->writeEntry((const char *)key , list);

	list.clear();
      }

}

void CubeBox::restoreGame(KConfig *config)
{
  QStrList list;
  list.setAutoDelete(TRUE);
  QString owner, value, key;
  
  for(int row=0; row < cubeDim ; row++)
    for(int column=0; column < cubeDim ; column++)
      {
	key.sprintf("%u,%u",row,column);
	config->readListEntry((const char *) key, list);
	owner=list.first();
	value=list.next();
	cubes[row][column]->setOwner(owner.toInt());
	cubes[row][column]->setValue(value.toInt());
	
	list.clear();
      }
      
     
  // set aktual player
  int onTurn=config->readNumEntry("onTurn",1);
  actualPlayer=onTurn;
  emit playerChanged(onTurn);
}


//#include "cubebox.moc"


